import pandas as pd
import numpy as np
import glob
import os
import os.path as path
from matplotlib import pyplot as plt
from glob import glob
import mne
import scipy
from scipy import stats
from scipy import fft
from tqdm import tqdm_notebook
import eeglib
from multiprocessing import Pool
import multiprocessing
import time
import pickle


power_band = {'alpha': (8, 13), 'beta': (13, 35), 'delta': (1, 4), 'gamma': (35, 50), 'theta': (4, 8)}

class Utilities:
    def save_object(obj, filename):
        with open(filename, 'wb') as outp:  # Overwrites any existing file.
            pickle.dump(obj, outp, pickle.HIGHEST_PROTOCOL)
            
    def pickle_loader(filename):
        """ Deserialize a file of pickled objects. """
        with open(filename, "rb") as f:
            while True:
                try:
                    yield pickle.load(f)
                except EOFError:
                    break


import mne #for eeg (bdf) file
from matplotlib import pyplot as plt

def plot_channels(data : mne.io.edf.edf.RawEDF,interval_min):
    ax = [plt.subplot(3,2,i) for i in range(1,7)]
    j = interval_min
    for i in range(6):
        if data._data[j,1:2]:
            ax[i].plot(data._data[j,:])
            ax[i].set_title(data.ch_names[j])
            j = j+1

def plot3Inter(raw1:mne.io.edf.edf.RawEDF,raw2:mne.io.edf.edf.RawEDF,raw3:mne.io.edf.edf.RawEDF,id_channel:int,min_val:int,max_val:int):
    ax1 = plt.subplot(2,2,1)
    ax2 = plt.subplot(2,2,2)
    ax3 = plt.subplot(2,2,3)

    ax1.set_title('Healthy - ' + raw1.ch_names[id_channel])
    ax1.plot(raw1._data[id_channel,min_val:max_val])

    ax2.set_title('Stroke - ' +raw2.ch_names[id_channel])
    ax2.plot(raw2._data[id_channel,min_val:max_val])

    ax3.set_title('TIA - ' +raw3.ch_names[id_channel])
    ax3.plot(raw3._data[id_channel,min_val:max_val])

def to_minutes(time:str):
    h = int(time.split(':')[0])
    m = int(time.split(':')[1])

    final_time = h*60 + m

    return final_time

def plotMultiEEGInter(raw1:mne.io.edf.edf.RawEDF,raw2:mne.io.edf.edf.RawEDF,raw3:mne.io.edf.edf.RawEDF,id_channel:int,min_val:int,max_val:int,total_plot:int):
    nbcols = total_plot // 2 + 1
    ax = [plt.subplot(nbcols,2,i) for i in range(1,nbcols+1)]

    l = 1
    for i in range(nbcols * 2):
        ax[l].set_title(raw1.ch_names[id_channel])
        ax[l].plot(raw1._data[id_channel,min_val:max_val])

        l += l+1
        ax[l].set_title(raw2.ch_names[id_channel])
        ax[l].plot(raw2._data[id_channel,min_val:max_val])

        l += l+1
        ax[l].set_title(raw3.ch_names[id_channel])
        ax[l].plot(raw3._data[id_channel,min_val:max_val])


import pandas as pd
import numpy as np
import glob
import os
import os.path as path
from matplotlib import pyplot as plt
from glob import glob
import mne
import scipy
from scipy import stats
from scipy import fft
from tqdm import tqdm_notebook
import eeglib
from multiprocessing import Pool
import multiprocessing
import time
import pickle


power_band = {'alpha': (8, 13), 'beta': (13, 35), 'delta': (1, 4), 'gamma': (35, 50), 'theta': (4, 8)}


class Features_extraction:
    
    
    def sampEns(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.sampEn(data[i]))

        return np.asarray(list)
    
    def mean(data):
        return np.mean(data,axis=-1)

    def std(data):
        return np.std(data,axis=-1)

    def var(data):
            return np.var(data,axis=-1)

    def minim(data):
            return np.min(data,axis=-1)

    def maxim(data):
            return np.max(data,axis=-1)

    def argminim(data):
            return np.argmin(data,axis=-1)

    def argmaxim(data):
            return np.argmax(data,axis=-1)

    def mean_square(data):
            return np.mean(data**2,axis=-1)

    def rms(data): #root mean square
            return  np.sqrt(np.mean(data**2,axis=-1))

    def dasd(data): #difference absolute standard deviation
        dasd = np.sqrt(np.mean(np.square(np.diff(data,axis=-1)),axis=-1)) 
        return dasd

    def abs_diffs_signal(data):
        return np.sum(np.abs(np.diff(data,axis=-1)),axis=-1)

    def kurtosis(data): 
            return  stats.kurtosis(data,axis=-1)

    def skews(data): 
            return  stats.skew(data,axis=-1)

    def log_detec(data):
        return np.exp(np.mean(np.log(np.abs(data)),axis=-1))

    def PFDs(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.PFD(data[i]))
        return np.asarray(list)

    def DFAs(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.DFA(data[i]))
        return np.asarray(list)

    def LZCs(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.LZC(data[i]))
        return np.asarray(list)

    def HFDs(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.HFD(data[i]))
        return np.asarray(list)

    def countSignChanges(data):
        list = []
        for i in range(len(data)):
            list.append(eeglib.features.countSignChanges(data[i]))
        return np.asarray(list)

    def hjorthActivities(helper:eeglib.helpers.EDFHelper):
        return helper.eeg.hjorthActivity()

    def hjorthComplexities(helper:eeglib.helpers.EDFHelper):
        return helper.eeg.hjorthComplexity()

    def hjorthMobilities(helper:eeglib.helpers.EDFHelper):
        return helper.eeg.hjorthMobility()

    def sampEns(helper:eeglib.helpers.EDFHelper):
        return helper.eeg.sampEn()

    def DTWs(helper:eeglib.helpers.EDFHelper):
        return helper.eeg.DTW()

    def bandsPower(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        dp = helper_stroke.eeg.bandPower(bands = power_band)
        df = pd.DataFrame(bp)
        return dp,df

    def alphaBand(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        return np.asarray(bandsPower(helper)[1].iloc[:,0].to_list())

    def betaBand(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        return np.asarray(bandsPower(helper)[1].iloc[:,1].to_list())

    def deltaBand(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        return np.asarray(bandsPower(helper)[1].iloc[:,1].to_list())

    def gammaBand(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        return np.asarray(bandsPower(helper)[1].iloc[:,3].to_list())

    def thetaBand(helper:eeglib.helpers.EDFHelper,powerBands = power_band):
        return np.asarray(bandsPower(helper)[1].iloc[:,4].to_list())


    def features_extraction(data,helper:eeglib.helpers.EDFHelper):
        features_dict = {
            'mean'
        }
        return [
            mean(data),std(data),var(data),minim(data),maxim(data),argminim(data),
            argmaxim(data),mean_square(data),rms(data),
            dasd(data),abs_diffs_signal(data),kurtosis(data),skews(data),
            log_detec(data),PFDs(data),DFAs(data),LZCs(data),
            HFDs(data),countSignChanges(data),hjorthActivities(helper),
            hjorthComplexities(helper),hjorthMobilities(helper),
            sampEns(helper),alphaBand(helper),betaBand(helper),deltaBand(helper),
            gammaBand(helper),thetaBand(helper)
        ]

    def features_extract_to_dataframe(features:list,channels_names):

        list_features = []
        for j in range(len(channels_names)):
            dict_feat = {
                'channel':channels_names[j],
                'mean':features[0][j],
                'std':features[1][j],
                'var':features[2][j],
                'minim':features[3][j],
                'maxim':features[4][j],
                'argminim':features[5][j],
                'argmaxim':features[6][j],
                'mean_square':features[7][j],
                'rms':features[8][j],
                'dasd':features[9][j],
                'abs_diffs_signal':features[10][j],
                'kurtosis':features[11][j],
                'skews':features[12][j],
                'log_detec':features[13][j],
                'PFDs':features[14][j],
                'DFAs':features[15][j],
                'LZCs':features[16][j],
                'HFDs':features[17][j],
                'countSignChanges':features[18][j],
                'hjorthActivities':features[19][j],
                'hjorthComplexities':features[20][j],
                'hjorthMobilities':features[21][j],
                'sampEns':features[22][j],
                'alphaBand':features[23][j],
                'betaBand':features[24][j],
                'deltaBand':features[25][j],
                'gammaBand':features[26][j],
                'thetaBand':features[27][j],
            }
            #print(channels_names[j])
            list_features.append(dict_feat)

        return pd.DataFrame(list_features)


import pandas as pd
import numpy as np
import glob
import os
import os.path as path
from matplotlib import pyplot as plt
from glob import glob
import mne
import scipy
from scipy import stats
from scipy import fft
from tqdm import tqdm_notebook
import eeglib
from multiprocessing import Pool
import multiprocessing
import time
import pickle


power_band = {'alpha': (8, 13), 'beta': (13, 35), 'delta': (1, 4), 'gamma': (35, 50), 'theta': (4, 8)}



        
class Preprocessing:
    #def __init__(self):
    
    #channels_names = ['F7','Fp1','Fp2','F8','F3','Fz','F4','C3','Cz','P8','P7','Pz','P4','T3','P3','O1','O2','C4','T4','A2']
    
    def features_extraction(data : np.ndarray, channel_name : str):
        mean = np.mean(data)
        std = np.std(data)
        var = np.var(data)
        rms = np.sqrt(np.mean(data**2))
        ptp = np.ptp(data)
        minim = np.min(data,axis=-1)
        maxim = np.max(data,axis=-1)
        argminim = np.argmin(data,axis=-1)
        argmaxim = np.argmax(data,axis=-1)
        dasd = np.sqrt(np.mean(np.square(np.diff(data,axis=-1)),axis=-1)) #difference absolute standard deviation
        PFD = eeglib.features.PFD(data)
        DFA = eeglib.features.DFA(data)
        HFD = eeglib.features.HFD(data)
        LZC = eeglib.features.LZC(data)
        kurtosis = stats.kurtosis(data)
        skew = stats.skew(data)
        countSignChanges = eeglib.features.countSignChanges(data)
        hjorthActivity = eeglib.features.hjorthActivity(data)
        hjorthComplexity = eeglib.features.hjorthComplexity(data)
        hjorthMobility = eeglib.features.hjorthMobility(data)
        sampEn = eeglib.features.sampEn(data)


        features = {
            channel_name+'_mean': mean,
            channel_name+'_std': std,
            channel_name+'_var': var,
            channel_name+'_rms': rms,
            channel_name+'_ptp': ptp,
            channel_name+'_minim': minim,
            channel_name+'_maxim': maxim,
            channel_name+'_argminim': argminim,
            channel_name+'_argmaxim': argmaxim, 
            channel_name+'_dasd': dasd,
            channel_name+'_PFD': PFD,
            channel_name+'_DFA': DFA,
            channel_name+'_HFD': HFD,
            channel_name+'_LZC': LZC,
            channel_name+'_kurtosis': kurtosis,
            channel_name+'_skew': skew,
            channel_name+'_countSignChanges': countSignChanges,
            channel_name+'_hjorthActivity' : hjorthActivity,
            channel_name+'_hjorthComplexity' : hjorthComplexity,
            channel_name+'_hjorthMobility' : hjorthMobility,
            channel_name+'_sampEn' : sampEn
        }
        return features       
    
    def add_eog(raw, channels, new_name):
        """Computes a single bipolar EOG channel from a list of possible names."""

        # Check that exactly two of the provided channels are in the data
        channels = [ch for ch in channels if ch in raw.ch_names]
        assert len(channels) == 2, (
            'Could not find exactly two channels for computing bipolar '
            f'{new_name}. Please provide different channel names or choose `None`')

        # Compute bipolar EOG channel
        anode = channels[0]
        cathode = channels[1]
        print(f'Adding bipolar channel {new_name} ({anode} - {cathode})')
        raw = set_bipolar_reference(
            raw, anode, cathode, new_name, drop_refs=False, verbose=False)
        raw = raw.set_channel_types({new_name: 'eog'})

        return raw

    def interpolate_bad_channels(raw, bad_channels=None, auto_bad_channels=None):
        """Interpolates any channels from the two lists."""

        # Combine lists of bad channels
        all_bad_channels = []
        if bad_channels is not None and bad_channels != 'auto':
            all_bad_channels += bad_channels
        if auto_bad_channels is not None:
            all_bad_channels += auto_bad_channels

        # Interpolate bad channels
        if all_bad_channels != []:
            raw.info['bads'] += auto_bad_channels
            raw = raw.interpolate_bads()

        return raw, all_bad_channels


    def correct_ica(raw, n_components=15, random_seed=1234, method='fastica'):
        """Corrects ocular artifacts using ICA and automatic component removal."""

        # Run ICA on a copy of the data
        raw_filt_ica = raw.copy()
        raw_filt_ica.load_data().filter(l_freq=1, h_freq=None, verbose=False)
        ica = ICA(
            n_components, random_state=random_seed, method=method)
        ica.fit(raw_filt_ica)

        # Remove bad components from the raw data
        eog_indices, _ = ica.find_bads_eog(
            raw, ch_name='VEOG', verbose=False)
        ica.exclude = eog_indices
        raw = ica.apply(raw)

        return raw, ica


    def read_data(file_path):
        raw = mne.io.read_raw_bdf(file_path,preload=True)
        raw.set_eeg_reference()
        raw.filter(l_freq=None,h_freq=45)
        epochs = mne.make_fixed_length_epochs(raw,duration=30,
                                              #overlap=0
                                             )
        epochs = epochs.get_data()

        return epochs

    def read_data_bdf(file_path):
        raw = mne.io.read_raw_bdf(file_path,preload=True)
        raw.set_eeg_reference()
        raw.filter(l_freq=None,h_freq=45)

        edf_data = raw.get_data()

        return edf_data

    def eeg_preprocessing(file_path,epochs_duration=20,l_freq=.5,h_freq=45,plot=False):
        no_eeg = ['ACC0','ACC1','ACC2','Packet Counter','TRIGGER']
        mont1020 = create_montage_10_20(channels_names)

        raw = mne.io.read_raw_bdf(file_path,preload=True)
        raw = add_eog(raw_stroke,['Fp1','Fp2'],'VEOG')
        raw = raw.set_montage(mont1020,on_missing = 'ignore')
        raw, all_bad_channels = interpolate_bad_channels(raw)
        raw, ica = correct_ica(raw.copy().drop_channels(no_eeg))
        raw.filter(l_freq=.5,h_freq=45)

        if plot:
            print('Ploting PSD')
            raw.plot_psd()
            print('Components of ICA')
            ica.plot_components(outlines = 'skirt')
            print('Ploting Raw')
            raw.plot()
            print('Effect of ICA in the Raw Data')
            ica.plot_sources(raw);

        epochs = mne.make_fixed_length_epochs(raw,duration=epochs_duration,
                                              #overlap=0
                                             )

        return raw,ica,epochs

    


    def features_extraction_eeg(data : np.ndarray,
                                channels_names:list = ['F7','Fp1','Fp2','F8','F3','Fz','F4','C3'
                                                       ,'Cz','P8','P7','Pz','P4','T3','P3','O1',
                                                       'O2','C4','T4','A2'], 
                                number_of_channels:int = 20):
        i = 0
        features_channel_list = []
        dall = {}
        
        if len(data)>=number_of_channels:
            for i in range(number_of_channels):
                print("Extraction of features's channels "+channels_names[i] +" in progress...")
                features = features_extraction(data=data[i],channel_name=channels_names[i])
                features_channel_list.append(features)

                print('features of channels '+channels_names[i]+' well extracted')
            print('----- All Done Well -----')
            for i in features_channel_list:
                dall.update(i)
        else:
            print("Error: The number of channels specified it's too many")

        return dall
    

    def features_extraction_eeg_helper(helper:eeglib.helpers.EDFHelper,
                                channels_names:list = ['F7','Fp1','Fp2','F8','F3','Fz','F4','C3'
                                                       ,'Cz','P8','P7','Pz','P4','T3','P3','O1',
                                                       'O2','C4','T4','A2'], 
                                number_of_channels:int = 20):
        i = 0
        features_channel_list = []
        dall = {}
        data = helper.data

        if len(data)>=number_of_channels:
            for i in range(number_of_channels):
                print("Extraction of features's channels "+channels_names[i] +" in progress...")
                features = features_extraction(data=data[i],channel_name=channels_names[i])
                features_channel_list.append(features)

                print('features of channels '+channels_names[i]+' well extracted')
            print('----- All Done Well -----')
            for i in features_channel_list:
                dall.update(i)
        else:
            print("Error: The number of channels specified it's too many")

        return dall

    def features_extraction_eeg_file(file_path):
        channels_names = ['F7','Fp1','Fp2','F8','F3','Fz','F4','C3','Cz','P8','P7','Pz','P4','T3','P3','O1','O2','C4','T4','A2']
        dataframe = pd.DataFrame()
        for i in range(len(file_path)):
            print("###### Features Extraction for the file " + str(i+1) + ' #######')
            helper = eeglib.helpers.EDFHelper(file_path[i],selectedSignals = channels_names,ICA=True)
            all_features = features_extraction_eeg(helper.data)
            dataframe = dataframe.append(all_features, ignore_index = True)

        return dataframe

    def features_extraction_eeg_file_parallel(file_path):
        channels_names  = ['F7','Fp1','Fp2','F8','F3','Fz','F4','C3','Cz','P8',
                           'P7','Pz','P4','T3','P3','O1','O2','C4','T4','A2']
        dataframe = pd.DataFrame()
        time.sleep(1)
        for i in range(len(file_path)):
            print("###### Features Extraction for the file " + str(i+1) + ' #######')
            helper = eeglib.helpers.EDFHelper(file_path[i],selectedSignals = channels_names,ICA=True)
            all_features = features_extraction_eeg(helper.data)
            dataframe = dataframe.append(all_features, ignore_index = True)

        return dataframe